/**
 * P4
 * http://snarfed.org/space/p4
 *
 * Copyright 2001-2004
 * Ken Ashcraft <kashcraft+p4@gmail.com>
 * Ryan Barrett <p4@ryanb.org>
 * Maulik Shah <p4@maulik.net>
 * Nathan Stoll <nathanstoll+p4@gmail.com>
 *
 * @file p4plugin.cpp
 *
 * Implements the plugin API calls listed in p4plugin.h.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */


#include "conn.h"
#include "receiveq.h"
#include "lka.h"
#include "msg_data.h"
#include "msg.h"
#include "debug.h"
#include "p4plugin.h"
#include "p4error.h"

#include <string>
using namespace std;


/*****************************************************************************
 * GLOBALS
 *****************************************************************************/
extern string glocal_user;


/*****************************************************************************
 * HELPER FUNCTIONS
 *****************************************************************************/
//  template <typename A, typename B = int, typename C = int, typename D = int,
//        typename E = int>
//  bool check_args(A a, B b = 1, C c = 1, D d = 1, E e = 1)
//  { return (a && b && c && d && e); }


// changed to multiple fns because fns can't have default template arguments :(
template <typename A>
inline bool check_args(A a)
{ return (a != 0); }

template <typename A, typename B>
inline bool check_args(A a, B b)
{ return (a && b); }

template <typename A, typename B, typename C>
inline bool check_args(A a, B b, C c)
{ return (a && b && c); }

template <typename A, typename B, typename C, typename D>
inline bool check_args(A a, B b, C c, D d)
{ return (a && b && c && d); }

template <typename A, typename B, typename C, typename D, typename E>
inline bool check_args(A a, B b, C c, D d, E e)
{ return (a && b && c && d && e); }


/*****************************************************************************
 * PLUGIN API FUNCTIONS
 *****************************************************************************/

// TODO: IMPLEMENT ME PLEASE! I'LL MAKE YOUR DREAMS COME TRUE!
uint p4_get_latency(const char *user, int *latency) {
  assert(0);
  return 0;
}



uint p4_broadcast(ushort plugin_id, const char *data, int size, bool encrypted)
{
  if (!check_args(plugin_id, data))
    return kerr_bad_arg;

  cmsg_data msg(plugin_id, data, size, encrypted);

  return msg.broadcast();
}


uint p4_send(ushort plugin_id, const char *user, const char *data, int size,
             bool encrypted)
{
  if (!check_args(plugin_id, user, data))
    return kerr_bad_arg;

  // is the user in the lka?
  string userstr(user);    // TODO: possible buffer overflow :(
  ulong addr;
  ushort port;
  uint ret = clka::instance().get_address(userstr.c_str(), addr, port);
  if (ret != kerr_success) {
    // user must not be connected...remove them from the LKA
    clka::instance().remove(user);
    return ret;
  }

  // send the message, and check the status of the connection that the message
  // was sent on
  cmsg_data msg(plugin_id, data, size, encrypted);

  const cconn &conn = msg.send(userstr);

  if (glocal_user == user)
    return kerr_success;

  ret = conn.wait_until_ready();
  if (ret != kerr_success) {
    // user must not be connected...remove them from the LKA
    clka::instance().remove(user);
    return ret;
  }

  assert(conn.get_status() == cconn::econnected);
  return kerr_success;
}


uint p4_get_num_plugins(const char *user, ushort *num_plugins)
{
  if (!check_args(user, num_plugins))
    return kerr_bad_arg;

  return clka::instance().get_num_plugins(user, *num_plugins);
}


uint p4_get_plugins(const char *user, ushort *plugin_ids, ushort num_plugins,
                    ushort *num_written)
{
  if (!check_args(user, plugin_ids, num_written) || num_plugins <= 0)
    return kerr_bad_arg;

  return clka::instance().get_plugins(user, plugin_ids, num_plugins,
                                      *num_written);
}


uint p4_get_num_users(ushort plugin_id, int *num_users)
{
  if (!check_args(num_users))
    return kerr_bad_arg;

  cdebug::instance() << acquire("lka_debug")
                     << "get_num_users(" << plugin_id << ")"
                     << release();

  uint ret, unum_users;
  if (plugin_id == 0)
    ret = clka::instance().get_num_users(unum_users);
  else
    ret = clka::instance().get_num_users(plugin_id, unum_users);

  *num_users = static_cast<int>(unum_users);
  return ret;
}


uint p4_get_users(ushort plugin_id, char **users, int num_users,
                  int *num_written)
{

  if (!check_args(users, num_written) || num_users <= 0)
    return kerr_bad_arg;

  cdebug::instance() << acquire("lka_debug")
                     << "get_users(" << plugin_id << ")"
                     << release();

  uint ret, uwrit = *num_written;
  if (plugin_id == 0)
    ret = clka::instance().get_users(users, num_users, uwrit);
  else
    ret = clka::instance().get_users(plugin_id, users, num_users, uwrit);

  *num_written = static_cast<int>(uwrit);
  return ret;
}


uint p4_authenticate(ushort plugin_id, bool *authenticated)
{
  // TODO: fill in real auth code
  *authenticated = false;
  return kerr_bad_sig;
}


uint p4_peek(ushort plugin_id, int *size)
{
  if (!check_args(plugin_id, size))
    return kerr_bad_arg;

  return creceiveq::instance().peek(plugin_id, *size);
}


uint p4_receive(ushort plugin_id, char *user, char *data, int size,
                int *size_written, bool *encrypted)
{
  if (!check_args(plugin_id, user, data, size_written, encrypted))
    return kerr_bad_arg;

  return creceiveq::instance().dequeue(plugin_id, user, data, size,
                                       *size_written, *encrypted);
}


